@@ -1,10 +1,9 @@
 /* -*-C-*-
 
-$Id: intprm.c,v 1.16 2008/01/30 20:02:13 cph Exp $
-
 Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
     1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
-    2006, 2007, 2008 Massachusetts Institute of Technology
+    2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
+    2017, 2018, 2019, 2020 Massachusetts Institute of Technology
 
 This file is part of MIT/GNU Scheme.
 
@@ -81,7 +80,29 @@ DEFINE_PRIMITIVE ("INTEGER-SUBTRACT", Prim_integer_subtract, 2, 2, 0)
      INTEGER_BINARY_OPERATION (integer_subtract)
 DEFINE_PRIMITIVE ("INTEGER-MULTIPLY", Prim_integer_multiply, 2, 2, 0)
      INTEGER_BINARY_OPERATION (integer_multiply)
-
+DEFINE_PRIMITIVE ("INTEGER-HAMMING-DISTANCE", Prim_integer_hamming_distance, 2, 2, 0)
+     INTEGER_BINARY_OPERATION (integer_hamming_distance)
+DEFINE_PRIMITIVE ("INTEGER-BITWISE-AND", Prim_integer_bitwise_and, 2, 2, 0)
+     INTEGER_BINARY_OPERATION (integer_bitwise_and)
+DEFINE_PRIMITIVE ("INTEGER-BITWISE-ANDC2", Prim_integer_bitwise_andc2, 2, 2, 0)
+     INTEGER_BINARY_OPERATION (integer_bitwise_andc2)
+DEFINE_PRIMITIVE ("INTEGER-BITWISE-ANDC1", Prim_integer_bitwise_andc1, 2, 2, 0)
+     INTEGER_BINARY_OPERATION (integer_bitwise_andc1)
+DEFINE_PRIMITIVE ("INTEGER-BITWISE-XOR", Prim_integer_bitwise_xor, 2, 2, 0)
+     INTEGER_BINARY_OPERATION (integer_bitwise_xor)
+DEFINE_PRIMITIVE ("INTEGER-BITWISE-IOR", Prim_integer_bitwise_ior, 2, 2, 0)
+     INTEGER_BINARY_OPERATION (integer_bitwise_ior)
+DEFINE_PRIMITIVE ("INTEGER-BITWISE-NOR", Prim_integer_bitwise_nor, 2, 2, 0)
+     INTEGER_BINARY_OPERATION (integer_bitwise_nor)
+DEFINE_PRIMITIVE ("INTEGER-BITWISE-EQV", Prim_integer_bitwise_eqv, 2, 2, 0)
+     INTEGER_BINARY_OPERATION (integer_bitwise_eqv)
+DEFINE_PRIMITIVE ("INTEGER-BITWISE-ORC2", Prim_integer_bitwise_orc2, 2, 2, 0)
+     INTEGER_BINARY_OPERATION (integer_bitwise_orc2)
+DEFINE_PRIMITIVE ("INTEGER-BITWISE-ORC1", Prim_integer_bitwise_orc1, 2, 2, 0)
+     INTEGER_BINARY_OPERATION (integer_bitwise_orc1)
+DEFINE_PRIMITIVE ("INTEGER-BITWISE-NAND", Prim_integer_bitwise_nand, 2, 2, 0)
+     INTEGER_BINARY_OPERATION (integer_bitwise_nand)
+
 #define INTEGER_UNARY_OPERATION(operator)				\
 {									\
   PRIMITIVE_HEADER (1);							\
@@ -97,6 +118,14 @@ DEFINE_PRIMITIVE ("INTEGER-SUBTRACT-1", Prim_integer_subtract_1, 1, 1, 0)
      INTEGER_UNARY_OPERATION (integer_subtract_1)
 DEFINE_PRIMITIVE ("INTEGER-LENGTH-IN-BITS", Prim_integer_length_in_bits, 1, 1, 0)
      INTEGER_UNARY_OPERATION (integer_length_in_bits)
+DEFINE_PRIMITIVE ("INTEGER-LENGTH", Prim_integer_length, 1, 1, 0)
+     INTEGER_UNARY_OPERATION (integer_length)
+DEFINE_PRIMITIVE ("INTEGER-FIRST-SET-BIT", Prim_integer_first_set_bit, 1, 1, 0)
+     INTEGER_UNARY_OPERATION (integer_first_set_bit)
+DEFINE_PRIMITIVE ("INTEGER-BIT-COUNT", Prim_integer_bit_count, 1, 1, 0)
+     INTEGER_UNARY_OPERATION (integer_bit_count)
+DEFINE_PRIMITIVE ("INTEGER-BITWISE-NOT", Prim_integer_bitwise_not, 1, 1, 0)
+     INTEGER_UNARY_OPERATION (integer_bitwise_not)
 
 DEFINE_PRIMITIVE ("INTEGER-DIVIDE", Prim_integer_divide, 2, 2, 0)
 {
@@ -144,20 +173,19 @@ DEFINE_PRIMITIVE ("INTEGER->FLONUM", Prim_integer_to_flonum, 2, 2, 0)
     SCHEME_OBJECT integer = (ARG_REF (1));
     long control = (arg_index_integer (2, 4));
     if (FIXNUM_P (integer))
-    {
-      long X = (FIXNUM_TO_LONG (integer));
-      double Y = ((double) X);
-
-      if (((long) Y) == X)
-	PRIMITIVE_RETURN (FIXNUM_TO_FLONUM (integer));
-      if ((control & 2) != 0)
-	error_bad_range_arg (1);
-      PRIMITIVE_RETURN (SHARP_F);
-    }
-    if (bignum_fits_in_word_p
-	(integer,
-	 (((control & 1) != 0) ? DBL_MANT_DIG : DBL_MAX_EXP),
-	 0))
+      {
+	double d = (FIXNUM_TO_DOUBLE (integer));
+	if ((0 == (control & 1))
+	    || ((DOUBLE_TO_FIXNUM_P (d))
+		&& (integer == (DOUBLE_TO_FIXNUM (d)))))
+	  PRIMITIVE_RETURN (double_to_flonum (d));
+	if ((control & 2) != 0)
+	  error_bad_range_arg (1);
+	PRIMITIVE_RETURN (SHARP_F);
+      }
+    if (((control & 1) != 0)
+	? (LOSSLESS_BIGNUM_TO_DOUBLE_P (integer))
+	: (BIGNUM_TO_DOUBLE_P (integer)))
       PRIMITIVE_RETURN (BIGNUM_TO_FLONUM (integer));
     if ((control & 2) != 0)
       error_bad_range_arg (1);
@@ -165,18 +193,54 @@ DEFINE_PRIMITIVE ("INTEGER->FLONUM", Prim_integer_to_flonum, 2, 2, 0)
   }
 }
 
+DEFINE_PRIMITIVE ("INTEGER-NONNEGATIVE-ONE-BITS", Prim_integer_nonnegative_one_bits, 2, 2, 0)
+{
+  PRIMITIVE_HEADER (2);
+  PRIMITIVE_RETURN
+    (integer_nonnegative_one_bits
+     ((arg_ulong_integer (1)), (arg_ulong_integer (2))));
+}
+
+DEFINE_PRIMITIVE ("INTEGER-NEGATIVE-ZERO-BITS", Prim_integer_negative_zero_bits, 2, 2, 0)
+{
+  PRIMITIVE_HEADER (2);
+  PRIMITIVE_RETURN
+    (integer_negative_zero_bits
+     ((arg_ulong_integer (1)), (arg_ulong_integer (2))));
+}
+
 DEFINE_PRIMITIVE ("INTEGER-SHIFT-LEFT", Prim_integer_shift_left, 2, 2, 0)
+{
+  PRIMITIVE_HEADER (2);
+  CHECK_ARG (1, INTEGER_P);
+  PRIMITIVE_RETURN
+    (integer_shift_left ((ARG_REF (1)), (arg_ulong_integer (2))));
+}
+
+DEFINE_PRIMITIVE ("INTEGER-SHIFT-RIGHT", Prim_integer_shift_right, 2, 2, 0)
 {
   PRIMITIVE_HEADER (2);
   CHECK_ARG (1, INTEGER_P);
   {
     SCHEME_OBJECT n = (ARG_REF (1));
-    if (integer_negative_p (n))
-      error_bad_range_arg (1);
-    PRIMITIVE_RETURN (integer_shift_left (n, (arg_ulong_integer (2))));
+    SCHEME_OBJECT m = (ARG_REF (2));
+    if (FIXNUM_P (m))
+      {
+	if (FIXNUM_NEGATIVE_P (m))
+	  error_bad_range_arg (2);
+	PRIMITIVE_RETURN (integer_shift_right (n, (FIXNUM_TO_ULONG (m))));
+      }
+    else if (BIGNUM_P (m))
+      {
+	if (BIGNUM_NEGATIVE_P (m))
+	  error_bad_range_arg (2);
+	PRIMITIVE_RETURN (FIXNUM_ZERO);
+      }
+    else
+      error_wrong_type_arg (2);
   }
 }
-
+
 static unsigned int
 list_to_integer_producer (void * context)
 {
