@@ -2,8 +2,8 @@
 
 Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
     1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
-    2006, 2007, 2008, 2009, 2010, 2011 Massachusetts Institute of
-    Technology
+    2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
+    2017, 2018, 2019, 2020 Massachusetts Institute of Technology
 
 This file is part of MIT/GNU Scheme.
 
@@ -29,15 +29,20 @@ USA.
 
 (declare (usual-integrations))
 
-(define (directory-processor input-type output-type process-file)
+(define (directory-processor input-type output-type process-file
+			     #!optional map-pathname)
   (let ((directory-read
 	 (let ((input-pattern
-		(make-pathname #f #f #f 'WILD input-type 'NEWEST)))
+		(make-pathname #f #f #f 'wild input-type 'newest)))
 	   (lambda (directory)
 	     (directory-read
 	      (merge-pathnames
 	       (pathname-as-directory (merge-pathnames directory))
-	       input-pattern))))))
+	       input-pattern)))))
+	(map-pathname
+	 (if (default-object? map-pathname)
+	     (lambda (pathname) pathname)
+	     map-pathname)))
     (lambda (input-directory #!optional output-directory force?)
       (let ((output-directory
 	     (if (default-object? output-directory) #f output-directory))
@@ -47,13 +52,14 @@ USA.
 		    (if (or force?
 			    (not (file-modification-time<=?
 				  (pathname-default-type pathname input-type)
-				  (let ((output-pathname
-					 (pathname-new-type pathname
-							    output-type)))
-				    (if output-directory
-					(merge-pathnames output-directory
-							 output-pathname)
-					output-pathname)))))
+				  (map-pathname
+				   (let ((output-pathname
+					  (pathname-new-type pathname
+							     output-type)))
+				     (if output-directory
+					 (merge-pathnames output-directory
+							  output-pathname)
+					 output-pathname))))))
 			(process-file pathname output-directory)))
 		  (if (pair? input-directory)
 		      (append-map! directory-read input-directory)
@@ -62,28 +68,13 @@ USA.
 (define sf-directory
   (directory-processor
    "scm"
-   (lambda () "bin")
+   (lambda () (if sf/cross-compiling? "nib" "bin"))
    (lambda (pathname output-directory)
-     (sf pathname output-directory))))
-
-(define compile-directory
-  (directory-processor
-   "bin"
-   (lambda ()
-     (compiler:compiled-code-pathname-type))
-   (lambda (pathname output-directory)
-     (compile-bin-file pathname output-directory))))
-
-(define sf-directory?)
-(define compile-directory?)
-(let ((show-pathname
-       (lambda (pathname output-directory)
-	 output-directory
-	 (newline)
-	 (write-string "Process file: ")
-	 (write-string (enough-namestring pathname)))))
-  (set! sf-directory? (directory-processor "scm" "bin" show-pathname))
-  (set! compile-directory? (directory-processor "bin" "com" show-pathname)))
+     (sf pathname output-directory))
+   (lambda (pathname)
+     (merge-pathnames
+      (enough-pathname (merge-pathnames pathname) sf/source-root)
+      sf/object-root))))
 
 (define (sf-conditionally filename #!optional echo-up-to-date?)
   (let ((kernel
