@@ -3,7 +3,7 @@
 Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
     1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
     2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
-    2017, 2018 Massachusetts Institute of Technology
+    2017, 2018, 2019, 2020 Massachusetts Institute of Technology
 
 This file is part of MIT/GNU Scheme.
 
@@ -73,30 +73,102 @@ USA.
 	     (assert-eqv (flo:rounding-mode) mode)))
 	 (assert-eqv (flo:rounding-mode) mode*))))))
 
+(define-syntax define-rounding-test
+  (syntax-rules ()
+    ((define-rounding-test name operator mode inputs outputs)
+     (define-test (symbol 'ROUNDING-MODE-INDEPENDENT ': mode '/ name)
+       (map (lambda (input output)
+              (lambda ()
+                (with-test-properties
+                    (lambda ()
+                      (assert-eqv
+                       (flo:with-rounding-mode mode
+                                               (lambda () (operator input)))
+                       output))
+                    'EXPRESSION `(,name ,input))))
+            inputs outputs)))))
+
 (for-each-rounding-mode
  (lambda (mode)
+   (define no-op identity-procedure)
    (define inputs '(-2.0 -1.5 -1.0 -0.5 -0.0 0.0 0.5 1.0 1.5 2.0))
-   (define (define-rounding-test name operator outputs)
-     (define-test (symbol 'ROUNDING-MODE-INDEPENDENT ': mode '/ name)
-       (lambda ()
-	 (do ((inputs inputs (cdr inputs))
-	      (outputs outputs (cdr outputs))
-	      (i 0 (+ i 1)))
-	     ((not (and (pair? inputs) (pair? outputs))))
-	   (let ((input (car inputs)) (output (car outputs)))
-	     (run-sub-test
-	      (lambda ()
-		(assert-eqv
-		 (flo:with-rounding-mode mode (lambda () (operator input)))
-		 output))))))))
-   (define-rounding-test 'CEILING ceiling
-     '(-2.0 -1.0 -1.0 -0.0 -0.0 0.0 1.0 1.0 2.0 2.0))
-   (define-rounding-test 'FLOOR floor
-     '(-2.0 -2.0 -1.0 -1.0 -0.0 0.0 0.0 1.0 1.0 2.0))
-   (define-rounding-test 'ROUND round
-     '(-2.0 -2.0 -1.0 -0.0 -0.0 0.0 0.0 1.0 2.0 2.0))
-   (define-rounding-test 'TRUNCATE truncate
-     '(-2.0 -1.0 -1.0 -0.0 -0.0 0.0 0.0 1.0 1.0 2.0))))
+   (define infs '(-inf.0 +inf.0))
+   ;; XXX Check NaNs without traps.
+   (let ((outputs '(-2.0 -1.0 -1.0 -0.0 -0.0 0.0 1.0 1.0 2.0 2.0)))
+     (define-rounding-test 'CEILING/INLINE ceiling mode inputs outputs)
+     (define-rounding-test 'CEILING/INLINE ceiling mode infs infs)
+     (define-rounding-test 'CEILING (no-op ceiling) mode inputs outputs)
+     (define-rounding-test 'CEILING (no-op ceiling) mode infs infs)
+     (define-rounding-test 'FLO:CEILING/INLINE flo:ceiling mode inputs outputs)
+     (define-rounding-test 'FLO:CEILING/INLINE flo:ceiling mode infs infs)
+     (define-rounding-test 'FLO:CEILING (no-op flo:ceiling) mode inputs
+       outputs)
+     (define-rounding-test 'FLO:CEILING (no-op flo:ceiling) mode infs infs)
+     (let ((outputs (map inexact->exact outputs)))
+       (define-rounding-test 'CEILING->EXACT/INLINE ceiling->exact mode inputs
+         outputs)
+       (define-rounding-test 'CEILING->EXACT (no-op ceiling->exact) mode inputs
+         outputs)
+       (define-rounding-test 'FLO:CEILING->EXACT/INLINE flo:ceiling->exact mode
+         inputs outputs)
+       (define-rounding-test 'FLO:CEILING->EXACT (no-op flo:ceiling->exact)
+         mode inputs outputs)))
+   (let ((outputs '(-2.0 -2.0 -1.0 -1.0 -0.0 0.0 0.0 1.0 1.0 2.0)))
+     (define-rounding-test 'FLOOR/INLINE floor mode inputs outputs)
+     (define-rounding-test 'FLOOR/INLINE floor mode infs infs)
+     (define-rounding-test 'FLOOR (no-op floor) mode inputs outputs)
+     (define-rounding-test 'FLOOR (no-op floor) mode infs infs)
+     (define-rounding-test 'FLO:FLOOR/INLINE flo:floor mode inputs outputs)
+     (define-rounding-test 'FLO:FLOOR/INLINE flo:floor mode infs infs)
+     (define-rounding-test 'FLO:FLOOR (no-op flo:floor) mode inputs outputs)
+     (define-rounding-test 'FLO:FLOOR (no-op flo:floor) mode infs infs)
+     (let ((outputs (map inexact->exact outputs)))
+       (define-rounding-test 'FLOOR->EXACT/INLINE floor->exact mode inputs
+         outputs)
+       (define-rounding-test 'FLOOR->EXACT (no-op floor->exact) mode inputs
+         outputs)
+       (define-rounding-test 'FLO:FLOOR->EXACT/INLINE flo:floor->exact mode
+         inputs outputs)
+       (define-rounding-test 'FLO:FLOOR->EXACT (no-op flo:floor->exact) mode
+         inputs outputs)))
+   (let ((outputs'(-2.0 -2.0 -1.0 -0.0 -0.0 0.0 0.0 1.0 2.0 2.0)))
+     (define-rounding-test 'ROUND/INLINE round mode inputs outputs)
+     (define-rounding-test 'ROUND/INLINE round mode infs infs)
+     (define-rounding-test 'ROUND (no-op round) mode inputs outputs)
+     (define-rounding-test 'ROUND (no-op round) mode infs infs)
+     (define-rounding-test 'FLO:ROUND/INLINE flo:round mode inputs outputs)
+     (define-rounding-test 'FLO:ROUND/INLINE flo:round mode infs infs)
+     (define-rounding-test 'FLO:ROUND (no-op flo:round) mode inputs outputs)
+     (define-rounding-test 'FLO:ROUND (no-op flo:round) mode infs infs)
+     (let ((outputs (map inexact->exact outputs)))
+       (define-rounding-test 'ROUND->EXACT/INLINE round->exact mode inputs
+         outputs)
+       (define-rounding-test 'ROUND->EXACT (no-op round->exact) mode inputs
+         outputs)
+       (define-rounding-test 'FLO:ROUND->EXACT/INLINE flo:round->exact mode
+         inputs outputs)
+       (define-rounding-test 'FLO:ROUND->EXACT (no-op flo:round->exact) mode
+         inputs outputs)))
+   (let ((outputs '(-2.0 -1.0 -1.0 -0.0 -0.0 0.0 0.0 1.0 1.0 2.0)))
+     (define-rounding-test 'TRUNCATE/INLINE truncate mode inputs outputs)
+     (define-rounding-test 'TRUNCATE/INLINE truncate mode infs infs)
+     (define-rounding-test 'TRUNCATE (no-op truncate) mode inputs outputs)
+     (define-rounding-test 'TRUNCATE (no-op truncate) mode infs infs)
+     (define-rounding-test 'FLO:TRUNCATE/INLINE flo:truncate mode inputs
+       outputs)
+     (define-rounding-test 'FLO:TRUNCATE/INLINE flo:truncate mode infs infs)
+     (define-rounding-test 'FLO:TRUNCATE (no-op flo:truncate) mode inputs
+       outputs)
+     (define-rounding-test 'FLO:TRUNCATE (no-op flo:truncate) mode infs infs)
+     (let ((outputs (map inexact->exact outputs)))
+       (define-rounding-test 'TRUNCATE->EXACT/INLINE truncate->exact mode
+         inputs outputs)
+       (define-rounding-test 'TRUNCATE->EXACT (no-op truncate->exact) mode
+         inputs outputs)
+       (define-rounding-test 'FLO:TRUNCATE->EXACT/INLINE flo:truncate->exact
+         mode inputs outputs)
+       (define-rounding-test 'FLO:TRUNCATE->EXACT (no-op flo:truncate->exact)
+         mode inputs outputs)))))
 
 ;++ Add tests for rounding-mode-dependent operations...
 
@@ -121,7 +193,7 @@ USA.
 		(cond ((assq elicitor-name elicitors)
 		       => (lambda (pair) (set-cdr! pair procedure)))
 		      (else
-		       (set-car! (list-tail descriptor 4)
+		       (set-car! (drop descriptor 4)
 				 (cons (cons elicitor-name procedure)
 				       elicitors)))))))
 	(else
@@ -220,6 +292,23 @@ USA.
       ;; (Note that if underflow is trapped, then the machine traps
       ;; whether or not the result is exact.  Go figure.)
       (flo:* (no-op .5000001) (flo:shift (no-op 1.) -1022)))))
+
+(if (not (zero? (flo:exception:subnormal-operand)))
+    (begin
+      (define-fpe-descriptor 'SUBNORMAL-OPERAND #t
+        flo:exception:subnormal-operand
+        ;; XXX Should have its own condition type, but this requires OS
+        ;; support for the siginfo code.
+	;; Changed to get test working on macOS:
+	;; condition-type:inexact-floating-point-result
+	condition-type:arithmetic-error)
+      (define-fpe-elicitor 'SUBNORMAL-OPERAND 'RAISE
+        (lambda ()
+          (flo:raise-exceptions! (flo:exception:subnormal-operand))))
+      (define-fpe-elicitor 'SUBNORMAL-OPERAND 'USE-A-SUBNORMAL-OPERAND
+        (let ((x (no-op flo:smallest-positive-subnormal)))
+          (lambda ()
+            (flo:+ x flo:smallest-positive-subnormal))))))
 
 (define (for-each-trappable-exception receiver)
   (for-each-exception
@@ -456,3 +545,52 @@ USA.
       (lambda ()
 	(assert-eqv (flo:trapped-exceptions)
 		    (flo:default-trapped-exceptions)))))
+
+(define-test 'preserving-environment
+  (lambda ()
+    ;; Trigger a bunch of floating-point exceptions _before_ we have
+    ;; registered interest in the floating-point environment -- but
+    ;; block thread-switching so nobody else can get in -- and then
+    ;; confirm that within flo:preserving-environment,
+    ;; flo:clear-exceptions! actually clears the exceptions.
+    (assert-eqv
+     (without-interruption
+      (lambda ()
+        ;; Trigger as many floating-point exceptions as we can.  If
+        ;; they trap, ignore it.  Don't enable floating-point traps
+        ;; since we're testing what happens if we haven't done any
+        ;; operations that touch the floating-point environment.
+        (ignore-errors
+         (lambda ()
+           (flo:sqrt -1.)
+           (flo:/ flo:smallest-positive-normal 2.)
+           (flo:sqrt 2.)
+           (flo:exp (flo:* 2. flo:greatest-normal-exponent-base-e))))
+        (flo:preserving-environment
+         (lambda ()
+           ;; Clear the exceptions.
+           (flo:clear-exceptions! (flo:supported-exceptions))
+           ;; Test the exceptions.  They should actually be cleared.
+           (flo:test-exceptions (flo:supported-exceptions))))))
+     0)))
+
+(define-test 'set-environment
+  (lambda ()
+    (define x0)
+    (define x1)
+    (define x2)
+    (define x3)
+    (flo:preserving-environment
+     (lambda ()
+       (flo:clear-exceptions! (flo:supported-exceptions))
+       (set! x0 (flo:test-exceptions (flo:supported-exceptions)))
+       (let ((env (flo:environment)))
+         (set! x1 (flo:test-exceptions (flo:supported-exceptions)))
+         (flo:raise-exceptions! (flo:exception:divide-by-zero))
+         (set! x2 (flo:test-exceptions (flo:supported-exceptions)))
+         (flo:set-environment! env)
+         (set! x3 (flo:test-exceptions (flo:supported-exceptions))))))
+    (assert-eqv x0 0)
+    (assert-eqv x1 0)
+    (assert-eqv x2 (flo:exception:divide-by-zero))
+    (assert-eqv x3 0)))
\ No newline at end of file
